# -*- coding: utf-8 -*-
# Copyright (C) 2006-2016 Mag. Christian Tanzer. All rights reserved
# Glasauergasse 32, A--1130 Wien, Austria. tanzer@swing.co.at
# ****************************************************************************
# This module is part of the package ATAX.
#
# This module is licensed under the terms of the BSD 3-Clause License
# <http://www.c-tanzer.at/license/bsd_3c.html>.
# ****************************************************************************
#
#++
# Name
#    ATAX.__init__
#
# Purpose
#    Initialize package `ATAX`
#
# Revision Dates
#    11-Feb-2006 (CT) Creation
#    11-Oct-2016 (CT) Add `__version__`
#    ««revision-date»»···
#--

from _TFL.Package_Namespace import Package_Namespace

__version__ = "1.1.2"

ATAX = Package_Namespace ()

del Package_Namespace

### __END__ ATAX.__init__
